/*
 * Copyright The OpenTelemetry Authors
 * SPDX-License-Identifier: Apache-2.0
 */

import io.opentelemetry.context.Context
import io.opentelemetry.instrumentation.api.tracer.BaseTracer
import io.opentelemetry.instrumentation.rxjava2.RxJava2AsyncSpanEndStrategy
import io.reactivex.Completable
import io.reactivex.Flowable
import io.reactivex.Maybe
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.observers.TestObserver
import io.reactivex.parallel.ParallelFlowable
import io.reactivex.processors.ReplayProcessor
import io.reactivex.processors.UnicastProcessor
import io.reactivex.subjects.CompletableSubject
import io.reactivex.subjects.MaybeSubject
import io.reactivex.subjects.ReplaySubject
import io.reactivex.subjects.SingleSubject
import io.reactivex.subjects.UnicastSubject
import io.reactivex.subscribers.TestSubscriber
import org.reactivestreams.Publisher
import org.reactivestreams.Subscriber
import org.reactivestreams.Subscription
import spock.lang.Specification

class RxJava2AsyncSpanEndStrategyTest extends Specification {
  BaseTracer tracer

  Context context

  def underTest = RxJava2AsyncSpanEndStrategy.INSTANCE

  void setup() {
    tracer = Mock()
    context = Mock()
  }

  static class CompletableTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Completable)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestObserver()

      when:
      def result = (Completable) underTest.end(tracer, context, Completable.complete())
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestObserver()

      when:
      def result = (Completable) underTest.end(tracer, context, Completable.error(exception))
      result.subscribe(observer)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span when completed"() {
      given:
      def source = CompletableSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Completable) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = CompletableSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Completable) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span once for multiple subscribers"() {
      given:
      def source = CompletableSubject.create()
      def observer1 = new TestObserver()
      def observer2 = new TestObserver()
      def observer3 = new TestObserver()

      when:
      def result = (Completable) underTest.end(tracer, context, source)
      result.subscribe(observer1)
      result.subscribe(observer2)
      result.subscribe(observer3)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer1.assertComplete()
      observer2.assertComplete()
      observer3.assertComplete()
    }
  }

  static class MaybeTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Maybe)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, Maybe.just("Value"))
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already empty"() {
      given:
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, Maybe.empty())
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, Maybe.error(exception))
      result.subscribe(observer)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span when completed"() {
      given:
      def source = MaybeSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onSuccess("Value")

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when empty"() {
      given:
      def source = MaybeSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = MaybeSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span once for multiple subscribers"() {
      given:
      def source = MaybeSubject.create()
      def observer1 = new TestObserver()
      def observer2 = new TestObserver()
      def observer3 = new TestObserver()

      when:
      def result = (Maybe<?>) underTest.end(tracer, context, source)
      result.subscribe(observer1)
      result.subscribe(observer2)
      result.subscribe(observer3)

      then:
      0 * tracer._

      when:
      source.onSuccess("Value")

      then:
      1 * tracer.end(context)
      observer1.assertValue("Value")
      observer1.assertComplete()
      observer2.assertValue("Value")
      observer2.assertComplete()
      observer3.assertValue("Value")
      observer3.assertComplete()
    }
  }

  static class SingleTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Single)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestObserver()

      when:
      def result = (Single<?>) underTest.end(tracer, context, Single.just("Value"))
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestObserver()

      when:
      def result = (Single<?>) underTest.end(tracer, context, Single.error(exception))
      result.subscribe(observer)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span when completed"() {
      given:
      def source = SingleSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Single<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onSuccess("Value")

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = SingleSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Single<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span once for multiple subscribers"() {
      given:
      def source = SingleSubject.create()
      def observer1 = new TestObserver()
      def observer2 = new TestObserver()
      def observer3 = new TestObserver()

      when:
      def result = (Single<?>) underTest.end(tracer, context, source)
      result.subscribe(observer1)
      result.subscribe(observer2)
      result.subscribe(observer3)

      then:
      0 * tracer._

      when:
      source.onSuccess("Value")

      then:
      1 * tracer.end(context)
      observer1.assertValue("Value")
      observer1.assertComplete()
      observer2.assertValue("Value")
      observer2.assertComplete()
      observer3.assertValue("Value")
      observer3.assertComplete()
    }
  }

  static class ObservableTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Observable)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestObserver()

      when:
      def result = (Observable<?>) underTest.end(tracer, context, Observable.just("Value"))
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestObserver()

      when:
      def result = (Observable<?>) underTest.end(tracer, context, Observable.error(exception))
      result.subscribe(observer)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span when completed"() {
      given:
      def source = UnicastSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Observable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = UnicastSubject.create()
      def observer = new TestObserver()

      when:
      def result = (Observable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span once for multiple subscribers"() {
      given:
      def source = ReplaySubject.create()
      def observer1 = new TestObserver()
      def observer2 = new TestObserver()
      def observer3 = new TestObserver()

      when:
      def result = (Observable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer1)
      result.subscribe(observer2)
      result.subscribe(observer3)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer1.assertComplete()
      observer2.assertComplete()
      observer3.assertComplete()
    }
  }

  static class FlowableTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Flowable)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, Flowable.just("Value"))
      result.subscribe(observer)

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, Flowable.error(exception))
      result.subscribe(observer)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span when completed"() {
      given:
      def source = UnicastProcessor.create()
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = UnicastProcessor.create()
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }

    def "ends span once for multiple subscribers"() {
      given:
      def source = ReplayProcessor.create()
      def observer1 = new TestSubscriber()
      def observer2 = new TestSubscriber()
      def observer3 = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer1)
      result.subscribe(observer2)
      result.subscribe(observer3)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer1.assertComplete()
      observer2.assertComplete()
      observer3.assertComplete()
    }
  }

  static class ParallelFlowableTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(ParallelFlowable)
    }

    def "ends span on already completed"() {
      given:
      def observer = new TestSubscriber()

      when:
      def result = (ParallelFlowable<?>) underTest.end(tracer, context, Flowable.just("Value").parallel())
      result.sequential().subscribe(observer)

      then:
      observer.assertComplete()
      1 * tracer.end(context)
    }

    def "ends span on already errored"() {
      given:
      def exception = new IllegalStateException()
      def observer = new TestSubscriber()

      when:
      def result = (ParallelFlowable<?>) underTest.end(tracer, context, Flowable.error(exception).parallel())
      result.sequential().subscribe(observer)

      then:
      observer.assertError(exception)
      1 * tracer.endExceptionally(context, exception)
    }

    def "ends span when completed"() {
      given:
      def source = UnicastProcessor.create()
      def observer = new TestSubscriber()

      when:
      def result = (ParallelFlowable<?>) underTest.end(tracer, context, source.parallel())
      result.sequential().subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      observer.assertComplete()
      1 * tracer.end(context)
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = UnicastProcessor.create()
      def observer = new TestSubscriber()

      when:
      def result = (ParallelFlowable<?>) underTest.end(tracer, context, source.parallel())
      result.sequential().subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      observer.assertError(exception)
      1 * tracer.endExceptionally(context, exception)
    }
  }

  static class PublisherTest extends RxJava2AsyncSpanEndStrategyTest {
    def "is supported"() {
      expect:
      underTest.supports(Publisher)
    }

    def "ends span when completed"() {
      given:
      def source = new CustomPublisher()
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onComplete()

      then:
      1 * tracer.end(context)
      observer.assertComplete()
    }

    def "ends span when errored"() {
      given:
      def exception = new IllegalStateException()
      def source = new CustomPublisher()
      def observer = new TestSubscriber()

      when:
      def result = (Flowable<?>) underTest.end(tracer, context, source)
      result.subscribe(observer)

      then:
      0 * tracer._

      when:
      source.onError(exception)

      then:
      1 * tracer.endExceptionally(context, exception)
      observer.assertError(exception)
    }
  }

  static class CustomPublisher implements Publisher<String>, Subscription {
    Subscriber<? super String> subscriber

    @Override
    void subscribe(Subscriber<? super String> subscriber) {
      this.subscriber = subscriber
      subscriber.onSubscribe(this)
    }

    def onComplete() {
      this.subscriber.onComplete()
    }

    def onError(Throwable exception) {
      this.subscriber.onError(exception)
    }

    @Override
    void request(long l) { }

    @Override
    void cancel() { }
  }
}
